// DTDGen.java: A SAX application to take an
// XML document and generate a DTD to which it conforms.
// 
// Author M. H. Kay
// revised 29 Apr 1998 to use SAX 1.0 interface
 
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;

import java.net.URL;
import java.util.*;
import java.io.*;

import BinaryTree;

public class DTDGen extends HandlerBase {

  BinaryTree elementList;
  Stack elementStack;

  /**
    * Entry point
    * Usage:  java -Dsax.parser=[parser-class] DTDGen input-file >output-file
    */

public static void main (String args[])
    throws java.lang.Exception
   {
				// Check the command-line arguments.
    if (args.length < 1 || args.length > 2) {
      System.err.println("Usage: java -Dsax.parser=[parser-class] DTDGen input-file >output-file");
      System.exit(1);
    } 

				// Create a new parser.
    Parser p = ParserFactory.makeParser();

                // Instantiate the application
    DTDGen app = new DTDGen();

				// Register the handler for basic document events.
    p.setDocumentHandler(app);

				// Run the parse.
    p.parse(new InputSource(new FileInputStream(args[0])));
 
  }

  public DTDGen () throws java.io.IOException
  {
    elementList = new BinaryTree();
    elementStack = new Stack();
  }

  
  /**
    * Handle the end of the document by printing the DTD
    */
  public void endDocument ()
  {

    // process the element types encountered, in turn

    Enumeration e=elementList.keys();
    while ( e.hasMoreElements() )
    {
        String elementname = (String) e.nextElement();
        ElementDetails ed = (ElementDetails) elementList.get(elementname);
 
        Vector children = ed.children;

        //EMPTY content
        if (children.isEmpty() && !ed.hasCharacterContent) 
            System.out.print("<!ELEMENT " + elementname + " EMPTY >\n");

        //CHARACTER content
        if (children.isEmpty() && ed.hasCharacterContent)
            System.out.print("<!ELEMENT " + elementname + " ( #PCDATA ) >\n");

        //ELEMENT content
        if (!(children.isEmpty()) && !ed.hasCharacterContent) {
          System.out.print("<!ELEMENT " + elementname + " ( ");
          Enumeration c=children.elements();
          while (true)
          {
            String child = (String) c.nextElement();
            System.out.print(child);
            if (c.hasMoreElements())
                System.out.print(" | ");
            else
                break;
          };
          System.out.print(" )* >\n");
        };

        //MIXED content
        if (!(children.isEmpty()) && ed.hasCharacterContent) {
          System.out.print("<!ELEMENT " + elementname + " ( #PCDATA");
          for (Enumeration c=children.elements(); c.hasMoreElements(); )
          {
            String child = (String) c.nextElement();
            System.out.print(" | " + child);
          };
          System.out.print(" )* >\n");
        };

        //Now examine the attributes encountered for this element type

        BinaryTree attlist = ed.attributes;
        Enumeration a=attlist.keys();
        while ( a.hasMoreElements() )
        {
            String attname = (String) a.nextElement();
            AttributeDetails ad = (AttributeDetails) attlist.get(attname);

            boolean required = (ad.occurrences==ed.occurrences);
            boolean isid = (ad.values.size()==ad.occurrences) && (ad.occurrences>10);
            boolean isenum = (ad.occurrences>10) && 
                             (ad.values.size()<=ad.occurrences/3) &&
                             (ad.values.size()<10);

            System.out.print("<!ATTLIST " + elementname + " " + attname + " ");

            if (isid) 
                System.out.print("ID");
            else if (isenum)
            {
                System.out.print("( ");
                Enumeration v=ad.values.keys();
                while (true)
                {   String val = (String) v.nextElement();
                    System.out.print(val);
                    if (v.hasMoreElements())
                       System.out.print(" | ");
                    else
                       break;
                };
                System.out.print(" )");
            }
            else
                System.out.print("CDATA");

            if (required)
                System.out.print(" #REQUIRED >\n");
            else if (ad.dflt!=null)
                System.out.print(" \"" + ad.dflt + "\" >\n");
            else
                System.out.print(" #IMPLIED >\n");
        };
        System.out.print("\n");
    };
   
  }

  /**
    * Handle the start of an element 
    */
  public void startElement (String name, AttributeList atts)
  {
        ElementDetails ed;
        ed = (ElementDetails) elementList.get(name);
        if (ed==null)  { 
            ed = new ElementDetails(name);
            elementList.put(name,ed);
        };

        // count occurrences of this element type

        ed.occurrences++;

        // handle the attributes accumulated for this element
        // add the new attribute list into the existing list for the element

        for (int i = 0; i < atts.getLength(); i++) {
            String att = atts.getName(i);
            String val = atts.getValue(i);
 
            AttributeDetails ad = (AttributeDetails) ed.attributes.get(att);
            if (ad==null) {
               ad=new AttributeDetails(att);
               ed.attributes.put(att, ad);
            };
 
            ad.values.put(val, Boolean.TRUE); //dummy value to indicate presence
            ad.occurrences++;
        };

        // now keep track of the nesting of elements

        if (!elementStack.isEmpty()) {
            ElementDetails parent = (ElementDetails) elementStack.peek();
            Vector children = parent.children;
            if (!children.contains(name)) children.addElement(name);
        };

        elementStack.push(ed); 
     
  }


  /**
    * Handle the end of an element 
    */
  public void endElement (String name)
  {
       elementStack.pop();
  }

  
  /**
    * Handle character data 
    */
  public void characters (char ch[], int start, int length)
  {
     String s = new String(ch, start, length);
     if (s.trim().equals("")) return;
     if (!elementStack.isEmpty()) {
        ElementDetails ed;
        ed = (ElementDetails) elementStack.peek();
        ed.hasCharacterContent = true;
     };

  }

}

class ElementDetails {
    String name;
    int occurrences;
    boolean hasCharacterContent;
    Vector children;
    BinaryTree attributes;

    public ElementDetails ( String name ) {
        this.name = name;
        this.occurrences = 0;
        this.hasCharacterContent = false;
        this.children = new Vector();
        this.attributes = new BinaryTree();
    }
       
}

class AttributeDetails {
    String name;
    String dflt;
    int occurrences;
    BinaryTree values; //used as a set

    public AttributeDetails ( String name ) {
        this.name = name;
        this.occurrences = 0;
        this.values = new BinaryTree();
    }
}


// end of DTDGen.java

